# -*- coding: utf-8 -*-
# Copyright 2019 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
""" Endpoint builder. """
from __future__ import absolute_import, unicode_literals

# stdlib imports
import textwrap

# 3rd party imports
from restible import ModelResource
from six import string_types        # pylint: disable=wrong-import-order

# local imports
from . import util


class EndpointBuilder(object):
    """ A helper class for building resource endpoints. """
    def __init__(self, route):
        self.route = route

    def actions_endpoint(self, actions):
        """ Extract spec for all action endpoints on the resource. """
        from restible import api_action

        res_cls = self.route.res_cls
        name = util.make_name(res_cls.name)
        endp = {}

        by_method = {}
        for action in actions:
            meta = api_action.get_meta(action)
            for method in meta.methods:
                by_method[method] = action

        for method, action in by_method.items():
            meta = api_action.get_meta(action)
            summary, desc = _parse_docstring(action)
            endp[method] = {
                "tags": [name],
                "summary": summary,
                "description": desc,
            }

            responses = self._get_responses(action, self_schema=meta.schema)
            if responses is not None:
                endp[method]['responses'] = responses

            if meta.schema:
                endp[method]['parameters'] = [{
                    "name": "payload",
                    "in": "body",
                    "schema": meta.schema,
                }]

        return endp

    def generic_endpoint(self):
        """ Extract spec for all generic endpoints on the resource. """
        res_cls = self.route.res_cls
        resource = res_cls()
        res_name = res_cls.name
        is_model_res = issubclass(res_cls, ModelResource)
        name = util.make_name(res_cls.name)
        endpoints = {}
        res_methods = {
            'create': 'create_item' if is_model_res else 'rest_create',
            'query': 'query_items' if is_model_res else 'rest_query',
        }

        if resource.implements('query'):
            res_method = getattr(res_cls, res_methods['query'])
            query_summary, query_desc = _parse_docstring(res_method)
            responses = self._get_responses(res_method, res_cls.schema, {
                "200": {
                    "description": "A list of {}s".format(res_name),
                    "schema": "__self_array__",
                },
                "401": util.RESPONSE_401
            })

            endpoints["get"] = {
                "tags": [name],
                "summary": query_summary,
                "description": query_desc,
                "responses": responses
            }

        if resource.implements('create'):
            res_method = getattr(res_cls, res_methods['create'])
            create_summary, create_desc = _parse_docstring(res_method)
            responses = self._get_responses(res_method, res_cls.schema, {
                "201": {
                    "description": "{} successfully created".format(name),
                    "schema": '__self__',
                },
                "401": util.RESPONSE_401
            })

            endpoints["post"] = {
                "tags": [name],
                "summary": create_summary,
                "description": create_desc,
                "parameters": [
                    {
                        "name": res_name,
                        "in": "body",
                        "description": "Initial {} data".format(res_name),
                        "schema": res_cls.schema,
                    }
                ],
                "responses": responses
            }

        return endpoints

    def detail_endpoint(self):
        """ Extract spec for all detail endpoints on the resource. """
        res_cls = self.route.res_cls
        resource = res_cls()
        res_name = res_cls.name
        is_model_res = issubclass(res_cls, ModelResource)
        name = util.make_name(res_cls.name)
        endpoints = {"parameters": res_cls.route_params}
        res_methods = {
            'get': 'get_item' if is_model_res else 'rest_get',
            'update': 'update_item' if is_model_res else 'rest_update',
            'delete': 'delete_item' if is_model_res else 'rest_delete',
        }

        if resource.implements('get'):
            res_method = getattr(res_cls, res_methods['get'])
            get_summary, get_desc = _parse_docstring(res_method)
            responses = self._get_responses(res_method, res_cls.schema, {
                "200": {
                    "description": "A list of {}s".format(res_name),
                    "schema": '__self_array__',
                },
                "401": util.RESPONSE_401
            })
            endpoints["get"] = {
                "tags": [name],
                "summary": get_summary,
                "description": get_desc,
                "responses": responses
            }

        if resource.implements('update'):
            res_method = getattr(res_cls, res_methods['update'])
            put_summary, put_desc = _parse_docstring(res_method)
            responses = self._get_responses(res_method, res_cls.schema, {
                "200": {
                    "description": "An updated {}".format(res_name),
                    "schema": '__self__',
                },
                "401": util.RESPONSE_401
            })
            endpoints["put"] = {
                "tags": [name],
                "summary": put_summary,
                "description": put_desc,
                "parameters": [
                    {
                        "name": res_name,
                        "in": "body",
                        "description": "{} data".format(res_name),
                        "schema": res_cls.schema,
                    }
                ],
                "responses": responses
            }

        if resource.implements('delete'):
            res_method = getattr(res_cls, res_methods['delete'])
            del_summary, del_desc = _parse_docstring(res_method)
            responses = self._get_responses(res_method, res_cls.schema, {
                "200": {"description": "Successfully deleted"},
                "401": util.RESPONSE_401
            })
            endpoints["delete"] = {
                "tags": [name],
                "summary": del_summary,
                "description": del_desc,
                "responses": responses,
            }

        return endpoints

    def _get_responses(self, handler, self_schema, defaults=None):
        if hasattr(handler, '_api_responses'):
            responses = handler._api_responses
        else:
            responses = defaults or {}

        for _, resp_spec in responses.items():
            schema = resp_spec.get('schema', None)

            if isinstance(schema, string_types):
                if schema == '__self__':
                    resp_spec['schema'] = self_schema
                elif schema == '__self_array__':
                    resp_spec['schema'] = {
                        "type": "array",
                        "items": self_schema
                    }

        return responses


def _parse_docstring(obj):
    if obj.__doc__ is None:
        return None, None

    docstring = obj.__doc__.strip()
    parts = docstring.split('\n\n', 1)
    summary = parts[0]

    if len(parts) == 2:
        desc = parts[1]
    else:
        desc = parts[0]

    desc = textwrap.dedent(desc)

    return summary, desc



# -*- coding: utf-8 -*-
# Copyright 2019 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
""" Various general purpose utilities. """
from __future__ import absolute_import, unicode_literals

# 3rd party imports
import yaml


RESPONSE_401 = {
    "description": "User needs to sign in",
    "schema": {
        "type": "object",
        "properties": {
            "detail": {
                "type": "string",
                "example": "Login required",
            }
        }
    }
}


RESPONSE_403 = {
    "description": "Access Denied",
    "schema": {
        "type": "object",
        "properties": {
            "detail": {
                "type": "string",
                "example": "Access Denied",
            }
        }
    }
}


RESPONSE_404 = {
    "description": "Not Found",
    "schema": {
        "type": "object",
        "properties": {
            "detail": {
                "type": "string",
                "example": "Not Found",
            }
        }
    }
}


RESPONSE_500 = {
    "description": "Server Error",
    "schema": {
        "type": "object",
        "properties": {
            "detail": {
                "type": "string",
                "example": "Server Error",
            }
        }
    }
}


def make_name(*words):
    """ Build a SnakeCase name out of words. """
    import itertools

    words = itertools.chain.from_iterable(w.split() for w in words)

    return ''.join(w.lower().capitalize() for w in words)


def yaml_dumps(data):
    """ Dump data as a YAML string. """
    return yaml.safe_dump(data, default_flow_style=False)


def yaml_write(data, path):
    """ Write data as YAML to the file under given path. """
    with open(path, 'w') as fp:
        fp.write(yaml_dumps(data))



# -*- coding: utf-8 -*-
# Copyright 2019 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
""" restible-swagger decorators. """
from __future__ import absolute_import, unicode_literals


from . import util


def responses(resp_def):
    """ Define responses for the given handlers. """
    def decorator(fn):  # pylint: disable=missing-docstring
        fn._api_responses = resp_def
        return fn
    return decorator


def response(status, response_def):
    """ Define a single response for the given handler.

    You can define multiple responses for the given handler by using this
    decorator multiple times.
    """
    def decorator(fn):  # pylint: disable=missing-docstring
        responses = getattr(fn, '_api_responses', {})
        responses[status] = response_def
        fn._api_responses = responses
        return fn
    return decorator


def response_200(description, array=False):
    """ A standard HTTP 200 response

    A quick helper to easily define a standard 200 response where the response
    schema matches the main resource schema for any given restible resource.
    """
    def decorator(fn):  # pylint: disable=missing-docstring
        responses = getattr(fn, '_api_responses', {})
        responses['200'] = {
            "description": description,
            "schema": "__self_array__" if array else "__self__",
        }
        fn._api_responses = responses
        return fn
    return decorator


def response_201(description):
    """ A standard HTTP 201 response

    A quick helper to easily define a standard 201 response where the response
    schema matches the main resource schema for any given restible resource.
    """
    def decorator(fn):  # pylint: disable=missing-docstring
        responses = getattr(fn, '_api_responses', {})
        responses['201'] = {
            "description": description,
            "schema": "__self__",
        }
        fn._api_responses = responses
        return fn
    return decorator


def response_401():
    """ A standard HTTP 401 response

    A quick helper for defining 401 responses. If you're using a custom error
    schema you'll have to build those manually. Otherwise you can use this
    little helper.
    """
    def decorator(fn):  # pylint: disable=missing-docstring
        responses = getattr(fn, '_api_responses', {})
        responses['401'] = util.RESPONSE_401
        fn._api_responses = responses
        return fn
    return decorator


def response_403():
    """ A standard HTTP 403 response

    A quick helper for defining 403 responses. If you're using a custom error
    schema you'll have to build those manually. Otherwise you can use this
    little helper.
    """
    def decorator(fn):  # pylint: disable=missing-docstring
        responses = getattr(fn, '_api_responses', {})
        responses['401'] = util.RESPONSE_401
        fn._api_responses = responses
        return fn
    return decorator


def response_404():
    """ A standard HTTP 404 response

    A quick helper for defining 404 responses. If you're using a custom error
    schema you'll have to build those manually. Otherwise you can use this
    little helper.
    """
    def decorator(fn):  # pylint: disable=missing-docstring
        responses = getattr(fn, '_api_responses', {})
        responses['404'] = util.RESPONSE_404
        fn._api_responses = responses
        return fn
    return decorator



# -*- coding: utf-8 -*-
# Copyright 2019 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
""" Main logic. """
from __future__ import absolute_import, unicode_literals

# stdlib imports
import os
import os.path

# 3rd party imports
from . import util
from .route import Route


def extract_api_spec(config, out_dir):
    """ Create OpenAPI spec using the given configuration. """

    if not os.path.exists(out_dir):
        os.makedirs(out_dir)

    for url, route_info in config.items():
        route = Route.load(url, route_info)
        route_spec = route.build_spec()
        res_file = os.path.join(out_dir, route.res_cls.name + '.yaml')

        print("-- \x1b[32mWriting \x1b[34m{}\x1b[0m".format(res_file))
        util.yaml_write(route_spec, res_file)



# -*- coding: utf-8 -*-
# Copyright 2019 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
""" Helper for defining API route mappings through config. """
from __future__ import absolute_import, unicode_literals

# stdlib imports
from typing import Any, Dict, List, Type

# 3rd party imports
import attr
from restible import RestResource

# local imports
from .endpoint import EndpointBuilder


@attr.s
class Route(object):
    """ Represents a single API route.

    A route is just a mapping of URL to resource that handles it. Each route
    will generate multiple URLs handled as each resource can handle generic and
    detail REST operation as well as all actions defined on the resource.
    """
    url = attr.ib(type=str)
    resource = attr.ib(type=str)
    route_params = attr.ib(type=str)
    actions = attr.ib(type=List[int])
    _res_cls = attr.ib(type=type, default=None)

    @classmethod
    def load(cls, url, route_info):
        """ Load route instance from configuration item. """
        return Route(
            url=url,
            resource=route_info['resource'],
            route_params=route_info['route_params'],
            actions=route_info.get('actions', []),
        )

    @property
    def res_cls(self):
        # type: () -> Type[RestResource]
        """ Return the resource class instance associated with this route. """
        if self._res_cls is None:
            mod_name, cls_name = self.resource.rsplit('.', 1)
            mod = __import__(mod_name, fromlist=[str(cls_name)])
            self._res_cls = getattr(mod, cls_name)

        return self._res_cls

    def build_spec(self):
        # type: () -> Dict[str, Any]
        """ Extract all OpenAPI specs for paths defined by the resource.

        Returns:
            Dict[str, Any]:
        """
        from restible import api_action

        builder = EndpointBuilder(self)

        list_path = self.res_cls.name
        detail_path = self.res_cls.name + '-detail'
        paths = {
            list_path: builder.generic_endpoint(),
            detail_path: builder.detail_endpoint(),
        }

        generic = []
        detail = []
        for action in self.res_cls().rest_actions():
            meta = api_action.get_meta(action)
            (generic if meta.generic else detail).append(action)

        actions_by_name = {}
        for action in detail:
            meta = api_action.get_meta(action)
            actions = actions_by_name.setdefault(meta.name, [])
            actions.append(action)

        for name, actions in actions_by_name.items():
            endp_name = '{res}-{action}'.format(res=self.res_cls.name, action=name)
            paths[endp_name] = builder.actions_endpoint(actions)

        return paths


# Used only in type hint comments
del Any, Dict, Type, RestResource
